/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dto.impl.lab;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonGetter;
import com.fasterxml.jackson.annotation.JsonProperty;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.Collection;
import java.util.Collections;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.Objects.max;
import static java.util.Comparator.comparing;

public class LabResultsImpl implements LabResults {

    private static final String PCR_RESULTS_PROPERTY = "pcrResults";
    private static final String ANTIGEN_RESULTS_PROPERTY = "antigenResults";
    private static final String CPR_PROPERTY = "cpr";

    @NonNull private final String cpr;
    @NonNull private final SortedSet<LabResult> results;
    @NonNull private final SortedSet<PcrResult> pcrResults;
    @NonNull private final SortedSet<AntigenResult> antigenResults;
    @Nullable private final LabResult mostRecentResult;
    @Nullable private final LabResult mostRecentPositiveResult;
    @Nullable private final LabResult mostRecentNegativeResult;
    @Nullable private final PcrResult mostRecentPcrResult;
    @Nullable private final PcrResult mostRecentPositivePcrResult;
    @Nullable private final PcrResult mostRecentNegativePcrResult;
    @Nullable private final AntigenResult mostRecentAntigenResult;
    @Nullable private final AntigenResult mostRecentPositiveAntigenResult;
    @Nullable private final AntigenResult mostRecentNegativeAntigenResult;

    @JsonCreator
    @SuppressWarnings({"java:S1117"})
    LabResultsImpl(
        @JsonProperty(CPR_PROPERTY) @NonNull String cpr,
        @JsonProperty(PCR_RESULTS_PROPERTY) @NonNull final Collection<? extends PcrResult> pcrResults,
        @JsonProperty(ANTIGEN_RESULTS_PROPERTY) @NonNull final Collection<? extends AntigenResult> antigenResults) {
        this.cpr = cpr;
        this.pcrResults = Collections.unmodifiableSortedSet(pcrResults.stream().collect(Collectors.toCollection(
            () -> new TreeSet<>(comparing(LabResult::resultDateTime).reversed()))));
        this.antigenResults = Collections.unmodifiableSortedSet(antigenResults.stream().collect(Collectors.toCollection(
            () -> new TreeSet<>(comparing(LabResult::resultDateTime).reversed()))));
        final SortedSet<LabResult> results = new TreeSet<>(comparing(LabResult::resultDateTime).reversed());
        results.addAll(pcrResults);
        results.addAll(antigenResults);
        this.results = Collections.unmodifiableSortedSet(results);
        mostRecentPcrResult = pcrResults.isEmpty() ? null : this.pcrResults.first();
        mostRecentAntigenResult = antigenResults.isEmpty() ? null : this.antigenResults.first();
        mostRecentPositivePcrResult = this.pcrResults.stream().filter(LabResult::positive).findFirst().orElse(null);
        mostRecentNegativePcrResult = this.pcrResults.stream().filter(LabResult::negative).findFirst().orElse(null);
        mostRecentPositiveAntigenResult = this.antigenResults.stream().filter(LabResult::positive).findFirst().orElse(null);
        mostRecentNegativeAntigenResult = this.antigenResults.stream().filter(LabResult::negative).findFirst().orElse(null);
        mostRecentResult = max(
            comparing(LabResult::resultDateTime),
            mostRecentPcrResult,
            mostRecentAntigenResult);
        mostRecentPositiveResult = max(
            comparing(LabResult::resultDateTime),
            mostRecentPositivePcrResult,
            mostRecentPositiveAntigenResult);
        mostRecentNegativeResult = max(
            comparing(LabResult::resultDateTime),
            mostRecentNegativePcrResult,
            mostRecentNegativeAntigenResult);
    }

    @Override
    @NonNull
    @JsonGetter
    @JsonProperty(CPR_PROPERTY)
    public String cpr() {
        return cpr;
    }

    @Override
    @NonNull
    public SortedSet<LabResult> results() {
        return results;
    }

    @Override
    @NonNull
    @JsonGetter
    @JsonProperty(PCR_RESULTS_PROPERTY)
    public SortedSet<PcrResult> pcrResults() {
        return pcrResults;
    }

    @NonNull
    @JsonGetter
    @JsonProperty(ANTIGEN_RESULTS_PROPERTY)
    public SortedSet<AntigenResult> antigenResults() {
        return antigenResults;
    }

    @Override
    @Nullable
    public LabResult mostRecentResult() {
        return mostRecentResult;
    }

    @Override
    @Nullable
    public LabResult mostRecentPositiveResult() {
        return mostRecentPositiveResult;
    }

    @Override
    @Nullable
    public LabResult mostRecentNegativeResult() {
        return mostRecentNegativeResult;
    }

    @Override
    @Nullable
    public PcrResult mostRecentPositivePcrResult() {
        return mostRecentPositivePcrResult;
    }

    @Override
    @Nullable
    public PcrResult mostRecentNegativePcrResult() {
        return mostRecentNegativePcrResult;
    }

    @Override
    @Nullable
    public AntigenResult mostRecentPositiveAntigenResult() {
        return mostRecentPositiveAntigenResult;
    }

    @Override
    @Nullable
    public AntigenResult mostRecentNegativeAntigenResult() {
        return mostRecentNegativeAntigenResult;
    }

    @Override
    @Nullable
    public PcrResult mostRecentPcrResult() {
        return mostRecentPcrResult;
    }

    @Override
    @Nullable
    public AntigenResult mostRecentAntigenResult() {
        return mostRecentAntigenResult;
    }
}
