/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.rules;

import dk.sundhedsdatastyrelsen.cbs.Objects;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.AntigenResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResult.PcrResult;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportDeniedEvaluation;
import dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvalution;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;

import static dk.sundhedsdatastyrelsen.cbs.Objects.firstNonNullOrDefaultSupplier;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvalution.deny;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvalution.grant;
import static dk.sundhedsdatastyrelsen.cbs.rules.evaluation.PassportEvalution.rejectedIndefinitely;
import static java.time.LocalDate.now;

public final class VaccinationValidator {

    private static final String NOT_COMPLETED_VACCINATION = "vaccination not completed";
    private static final String MINIMUM_DURATION_NOT_PASSED = "minimum duration not passed";
    private static final String VACCINATION_EXPIRED = "vaccination expired";
    private static final String POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT = "postive antigen result within suspension period without subsequent negative pcr result";
    private static final String POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD = "positive pcr result within suspension period";

    @NonNull private final Duration minimumDurationAfterVaccination;
    @NonNull private final Duration maximumDurationAfterVaccination;
    @NonNull private final Duration suspensionDuration;

    public VaccinationValidator(
        @NonNull final Duration minimumDurationAfterVaccination,
        @NonNull final Duration maximumDurationAfterVaccination,
        @NonNull final Duration suspensionDuration) {
        this.minimumDurationAfterVaccination = minimumDurationAfterVaccination;
        this.maximumDurationAfterVaccination = maximumDurationAfterVaccination;
        this.suspensionDuration = suspensionDuration;
    }

    /**
     * Evaluates vaccinations and lab results in the context of vaccination passport rules.
     * @return the result of the evaluation.
     */
    @NonNull
    @SuppressWarnings({"java:S2637", "ConstantConditions"})
    public final PassportEvalution evaluate(@NonNull final Vaccination vaccination, @NonNull final LabResults results) {
        final LocalDateTime completed = vaccination.completed();

        return firstNonNullOrDefaultSupplier(
            () -> grant(vaccinationValidUntil(completed)),
            // Reasons not to grant a vaccination passport follow below,
            // if none are met return above
            () -> notCompletedVaccination(completed),
            () -> minimumDurationAfterVaccinationHasNotPassed(completed),
            () -> maximumDurationAfterVaccinationHasPassed(completed),
            () -> suspendedDueToPositiveLabResult(results));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive lab results.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation suspendedDueToPositiveLabResult(@NonNull final LabResults results) {
        return Objects.firstNonNullSupplier(
            () -> thereIsAPositivePcrResultWithinSuspensionPeriod(results),
            () -> thereIsAPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(results));
    }

    /**
     * Evaluates if a passport should be denied temporarily due to positive antigen result within the suspension period
     * that has not be set aside by a subsequent negative pcr result.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation thereIsAPositiveAntigenResultWithinSuspensionPeriodThatHasNotBeenFollowedByANegativePcrResult(@NonNull final LabResults results) {
        final AntigenResult mostRecentPositiveAntigenResult = results.mostRecentPositiveAntigenResult();
        final PcrResult mostRecentNegativePcrResult = results.mostRecentNegativePcrResult();

        if(mostRecentPositiveAntigenResult == null) return null;
        if(resultOutsideSuspensionPeriod(mostRecentPositiveAntigenResult)) return null;
        if(pcrResultAfterAntigenResult(mostRecentNegativePcrResult, mostRecentPositiveAntigenResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositiveAntigenResult), POSITIVE_ANTIGEN_RESULT_WITHIN_SUSPENSION_PERIOD_WITHOUT_SUBSEQUENT_NEGATIVE_PCR_RESULT);
    }

    /**
     * Evaluates if a passport should be denied because a vaccination is not completed.
     * @param completed the completion date of the vaccination, may be null if the vaccination was not completed.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private static PassportDeniedEvaluation notCompletedVaccination(@Nullable final LocalDateTime completed) {
        return completed == null ? rejectedIndefinitely(NOT_COMPLETED_VACCINATION) : null;
    }

    /**
     * Evaluates if a passport should be denied because the minimum duration after a vaccination has been completed has
     * not passed.
     * @param completed the completion date of the vaccination.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation minimumDurationAfterVaccinationHasNotPassed(@NonNull final LocalDateTime completed) {
        final LocalDate vaccinationValidFrom = vaccinationValidFrom(completed);
        return now().isBefore(vaccinationValidFrom) ? deny(vaccinationValidFrom, MINIMUM_DURATION_NOT_PASSED): null;
    }

    /**
     * Evaluates if a passport should be denied because the maximum duration after a vaccination has been completed has
     * not passed.
     * @param completed the completion date of the vaccination.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation maximumDurationAfterVaccinationHasPassed(@NonNull final LocalDateTime completed) {
        final LocalDate validUntil = vaccinationValidUntil(completed);
        return now().isAfter(validUntil) ? rejectedIndefinitely(VACCINATION_EXPIRED) : null;
    }

    /**
     * Evaluates if a passport should be denied because there is a positve pcr result within the suspension period.
     * @param results the lab results to look among.
     * @return the denied evaluation if the passport was denied, else <code>null</code>.
     */
    @Nullable
    private PassportDeniedEvaluation thereIsAPositivePcrResultWithinSuspensionPeriod(@NonNull final LabResults results) {
        final PcrResult mostRecentPositivePcrResult = results.mostRecentPositivePcrResult();

        if (mostRecentPositivePcrResult == null) return null;
        if (resultOutsideSuspensionPeriod(mostRecentPositivePcrResult)) return null;

        return deny(positiveResultExpiresSuspensionPeriod(mostRecentPositivePcrResult), POSITIVE_PCR_RESULT_WITHIN_SUSPENSION_PERIOD);
    }

    /**
     * Evaluates whether or not there a result is within the suspension period.
     * @param result the lab result to evaluate.
     * @return <code>false</code> if the result is within the suspension period, else <code>true</code>.
     */
    private boolean resultOutsideSuspensionPeriod(@NonNull final LabResult result) {
        final LocalDate resultTime = result.resultDateTime().toLocalDate();
        return resultTime.isBefore(startOfSuspensionPeriod());
    }

    /**
     * Evaluates whether or not an antigen result is before a pcr result.
     * @param pcrResult the most recent pcr result, can be <code>null</code>.
     * @param antigenResult the most recent positive antigen result.
     * @return <code>true</code> if the pcr result is not <code>null</code> and after the antigen result.
     */
    private static boolean pcrResultAfterAntigenResult(
        @Nullable final PcrResult pcrResult,
        @NonNull final AntigenResult antigenResult) {
        if(pcrResult == null) return false;

        final LocalDateTime dateTimeOfMostRecentPositiveAntigenResult = antigenResult.resultDateTime();
        final LocalDateTime dateTimeOfMostRecentNegativePcrResult = pcrResult.resultDateTime();
        return dateTimeOfMostRecentNegativePcrResult.isAfter(dateTimeOfMostRecentPositiveAntigenResult);
    }

    /**
     * Calculates the earliest date from which a completed vaccination can grant a passport
     * @param completed the completion date of the vaccination
     * @return the date
     */
    @NonNull
    private LocalDate vaccinationValidFrom(@NonNull LocalDateTime completed) {
        return completed.plus(minimumDurationAfterVaccination).toLocalDate();
    }

    /**
     * Calculates the latest date from which a completed vaccination can grant a passport
     * @param completed the completion date of the vaccination
     * @return the date
     */
    @NonNull
    private LocalDate vaccinationValidUntil(@NonNull final LocalDateTime completed) {
        return completed.plus(maximumDurationAfterVaccination).toLocalDate();
    }

    /**
     * Calculates the earliest day of a possible suspension of an completed vaccination due to positive results.
     * @return the date
     */
    @NonNull
    private LocalDate startOfSuspensionPeriod() {
        return now().minus(suspensionDuration);
    }

    /**
     * Calculates the last day of a possible suspension of an completed vaccination due to positive results.
     * @return the date
     */
    @NonNull
    private LocalDate positiveResultExpiresSuspensionPeriod(@NonNull final LabResult positiveResult) {
        return positiveResult.resultDateTime().plus(suspensionDuration).toLocalDate();
    }
}
